﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO.Ports;
using System.Linq;
using System.Net.Http;
using System.Security.Permissions;
using System.Text;
using System.Threading.Tasks;
using TelemetryIO.Models;

namespace TelemetryIO
{
    public class SerialHandler
    {
        public const int TELE_CMD_RD_ONCE =   1;
        private const int TELE_CMD_RD_MON_ON =   2;
        private const int TELE_CMD_RD_MON_OFF = 3;
        private const int TELE_CMD_RD_MON_ADD = 4;
        private const int TELE_CMD_RD_MON_REMOVE = 5;
        private const int TELE_CMD_RD_MON_REMOVEALL = 6;
        private const int TELE_CMD_WR =     10;
        private const int TELE_CMD_ABORT =     999;
        private const int TELE_CMD_HELLO =     9999;
        private const byte CRC8_POLYNOMIAL = 0x07;
        private const int CRC32_POLY = 0x04C11DB7;
        private const byte ESCAPE_BEGIN = 0xBE;
        private const byte ESCAPE_END = 0xED;
        private const byte ESCAPE_CHAR = 0x0E;
        private SerialPort com_port;
        static private List<byte> rx_buf = new List<byte>();
        static private byte[] data;
        public delegate void AnswerEventHandler<SerialEventArgs>(object sender, SerialEventArgs e);
        public event AnswerEventHandler<SerialEventArgs> AnswerReceived;
        private int int_test = 0;
        private float[] monitor = new float[32];
        public string view_str = "";
        Telemetry telemetry = Telemetry.Instance;
        

        public struct Telegram
        {
            public int cmd;
            public int slot;
            public int len;
            public int offset;
            public byte[] param;
        };

        public string PortName
        {
            get { return com_port.PortName; }
            set { com_port.PortName = value; }
        }

        public int BaudRate
        {
            get { return com_port.BaudRate; }
            set { com_port.BaudRate = value; }
        }

        public bool isOpen
        {
            get { return com_port.IsOpen; }
        }

        public SerialHandler()
        {
            com_port = new SerialPort();
        }

        public SerialHandler(string PortName, int BaudRate)
        {
            com_port = new SerialPort();
            com_port.PortName = PortName;
            com_port.BaudRate = BaudRate;
        }

        private void Serial_DataReceived(object sender, SerialDataReceivedEventArgs e)
        {
            SerialPort sp = (SerialPort)sender;
            

            while (sp.BytesToRead > 0)
            {
                if (!sp.IsOpen) return;
                byte b = (byte)sp.ReadByte();

                if (b == ESCAPE_END)
                {
                    byte[] unscape = EscapeSeqToBytes(rx_buf.ToArray());
                    uint checksum = crc32(unscape, unscape.Length - 4);
                    uint re_checksum = DataSlots.bytes_to_uint(unscape, unscape.Length - 4);
                    if (re_checksum == checksum)
                    {
                        //Parse telegram
                        int cmd = DataSlots.bytes_to_int(unscape, 0);
                        int slot = DataSlots.bytes_to_int(unscape, 4);
                        int len = DataSlots.bytes_to_int(unscape, 8);
                        int offset = DataSlots.bytes_to_int(unscape, 12);
                        byte[] data = new byte[len];
                        Array.Copy(unscape, 16, data, 0, len);

                        if (cmd == TELE_CMD_RD_ONCE)
                        {
                            DataSlots.setSlot(slot, data, offset, len);
                        }
                        else if (cmd == TELE_CMD_RD_MON_ON)
                        {
                            for(int i = 0; i < data.Length/4; i++)
                            {
                                monitor[i] = BitConverter.ToSingle(data, i*4);
                            }
                            view_str = string.Join(", ", monitor.Select(x => "" + x).ToArray(), 0, data.Length/4);
                        }
                        else if (cmd == TELE_CMD_WR)
                        {
                            if (data[0] == 'O' && data[1] == 'k')
                            {
                                Debug.Write("Data is written\n");
                                OnAnswerReceived("Data is written");
                            }
                        }else if(cmd == TELE_CMD_HELLO)
                        {
                            telemetry.setSlot(9999, data, 0, 0);
                            view_str = Encoding.ASCII.GetString(data);
                        }
                        
                    }
                }
                else if (b == ESCAPE_BEGIN)
                {
                    rx_buf.Clear();
                }
                else
                {
                    rx_buf.Add(b);
                }

            }
            
        }

        public void Open()
        {
            com_port.Open();
            com_port.DataReceived += Serial_DataReceived;
            byte[] to = prepareTelegram(TELE_CMD_HELLO, 0, new byte[0], 0);
            com_port.Write(to, 0, to.Length);
        }

        public void Close()
        {
            com_port.DataReceived -= Serial_DataReceived;
            com_port.Close();
        }

        byte crc8(byte[] data, int length)
        {
            byte crc = 0x00;  // Initial value

            for (int i = 0; i < length; i++)
            {
                crc ^= data[i];  // XOR with the data byte

                for (int j = 0; j < 8; j++)
                {
                    if ((byte)(crc & 0x80) > 0)
                    {
                        crc = (byte)((byte)(crc << 1) ^ CRC8_POLYNOMIAL);
                    }
                    else
                    {
                        crc = (byte)(crc << 1);
                    }
                }
            }

            return crc;
        }

        uint crc32(byte[] data, int length)
        {
            uint crc = 0xFFFFFFFF;  // Начальное значение CRC

            for (int i = 0; i < length; i++)
            {
                crc ^= (uint)data[i] << 24;  // XOR с текущим байтом

                for (int j = 0; j < 8; j++)
                {
                    if ((uint)(crc & 0x80000000) != 0)
                    {
                        crc = (crc << 1) ^ CRC32_POLY;
                    }
                    else
                    {
                        crc <<= 1;
                    }
                }
            }

            return crc;
        }

        public void sendTest()
        {
            //data - index in monitor
            byte[] to = prepareTelegram(TELE_CMD_RD_MON_ADD, 0, BitConverter.GetBytes(int_test), int_test, 4);
            com_port.Write(to, 0, to.Length);
            int_test++;
        }

        public void sendTestCont()
        {
            byte[] b = new byte[0];
            byte[] to = prepareTelegram(TELE_CMD_RD_MON_ON, 0, b, 0, 24);
            com_port.Write(to, 0, to.Length);
        }

        public void sendTestWrite(float[] f)
        {
            byte[] b = prepareTelegram(TELE_CMD_WR, 1000, f, 0);
            com_port.Write(b, 0, b.Length);
        }

        byte[] prepareTelegram<T>(int cmd, int slot, T[] load, int offset, int len = 0)
        {
            byte[] byteload = DataArrayToBytes(load);
            int total_len = 20 + byteload.Length;//cmd[4 bytes] + slot[4 bytes] + len[4 bytes] + offset[4 bytes] + load[len bytes]
            byte[] data = new byte[total_len];

            //Construct telegram
            data[0] = (byte)(0xFF & cmd);
            data[1] = (byte)(0xFF & (cmd >> 8));
            data[2] = (byte)(0xFF & (cmd >> 16));
            data[3] = (byte)(0xFF & (cmd >> 24));

            data[4] = (byte)(0xFF & slot);
            data[5] = (byte)(0xFF & (slot >> 8));
            data[6] = (byte)(0xFF & (slot >> 16));
            data[7] = (byte)(0xFF & (slot >> 24));

            int l = 0;
            if (cmd == TELE_CMD_WR) l = byteload.Length;
            else l = len;

            data[8] = (byte)(0xFF & l);
            data[9] = (byte)(0xFF & (l >> 8));
            data[10] = (byte)(0xFF & (l >> 16));
            data[11] = (byte)(0xFF & (l >> 24));

            data[12] = (byte)(0xFF & offset);
            data[13] = (byte)(0xFF & (offset >> 8));
            data[14] = (byte)(0xFF & (offset >> 16));
            data[15] = (byte)(0xFF & (offset >> 24));

            if(byteload.Length > 0)
            {
                //Copy data
                Array.Copy(byteload, 0, data, 16, byteload.Length);
            }
            //CRC32
            uint checksum = crc32(data, total_len - 4);
            data[total_len - 4] = (byte)(0xFF & checksum);
            data[total_len - 3] = (byte)(0xFF & (checksum >> 8));
            data[total_len - 2] = (byte)(0xFF & (checksum >> 16));
            data[total_len - 1] = (byte)(0xFF & (checksum >> 24));

            byte[] escape = BytesToEscapeSeq(data);
            byte[] ret = new byte[escape.Length + 2];
            Array.Copy(escape, 0, ret, 1, escape.Length);
            ret[0] = ESCAPE_BEGIN;
            ret[ret.Length - 1] = ESCAPE_END;
            return ret;
        }

        byte[] DataArrayToBytes<T>(T[] data)
        {
            if (data == null) return new byte[0];
            List<byte>ret = new List<byte> ();
            for(int i = 0; i < data.Length; i++)
            {
                if(typeof(T) == typeof(float))
                {
                    ret.AddRange(BitConverter.GetBytes((float)(object)data[i]));
                }else if(typeof(T) == typeof(int))
                {
                    ret.AddRange(BitConverter.GetBytes((int)(object)data[i]));
                }else if(typeof(T) == typeof(byte))
                {
                    ret.Add((byte)(object)data[i]);
                }else if(typeof(T) == typeof(bool))
                {
                    bool t = (bool)(object)data[i];
                    if (t) ret.Add(1);
                    else ret.Add(0);
                }
                
            }
            return ret.ToArray();
        }

        byte[] BytesToEscapeSeq(byte[] data)
        {
            List<byte> ret = new List<byte>();
            for (int i = 0; i < data.Length; i++)
            {
                if ((data[i] == ESCAPE_BEGIN) || (data[i] == ESCAPE_CHAR) || (data[i] == ESCAPE_END))
                {
                    ret.Add(ESCAPE_CHAR);
                    ret.Add((byte)(data[i] - 0x0E));
                }
                else
                {
                    ret.Add(data[i]);
                }
            }
            return ret.ToArray();
        }

        byte[] EscapeSeqToBytes(byte[] EscSeq)
        {
            List<byte> ret = new List<byte>();
            for (int i = 0; i < EscSeq.Length; i++)
            {
                if ((EscSeq[i] == ESCAPE_BEGIN) || (EscSeq[i] == ESCAPE_CHAR) || (EscSeq[i] == ESCAPE_END))
                {
                    i++;
                    ret.Add((byte)(EscSeq[i] + 0x0E));
                }
                else
                {
                    ret.Add(EscSeq[i]);
                }
            }
            return ret.ToArray();
        }

        protected virtual void OnAnswerReceived(string answer)
        {
            AnswerReceived?.Invoke(this, new SerialEventArgs(answer));
        }
    }

    public class SerialEventArgs: EventArgs
    {
        public string Answer { get; set; }

        public SerialEventArgs(string answer)
        {
            Answer = answer;
        }
    }
}
